@@ -31,7 +31,7 @@ from sys import maxsize
 from time import time
 
 class TMBasis():
-	def __init__(self, number_of_clauses, T, s, confidence_driven_updating=False, type_i_ii_ratio = 1.0, type_iii_feedback=False, focused_negative_sampling=False, output_balancing=False, d=200.0, platform='CPU', patch_dim=None, feature_negation=True, boost_true_positive_feedback=1, max_included_literals=None, number_of_state_bits_ta=8, number_of_state_bits_ind=8, weighted_clauses=False, clause_drop_p = 0.0, literal_drop_p = 0.0):
+	def __init__(self, number_of_clauses, T, s, confidence_driven_updating=False, type_i_ii_ratio = 1.0, type_iii_feedback=False, focused_negative_sampling=False, output_balancing=False, d=200.0, platform='CPU', patch_dim=None, feature_negation=True, boost_true_positive_feedback=1, max_included_literals=None, number_of_state_bits_ta=8, number_of_state_bits_ind=8, weighted_clauses=False, clause_drop_p = 0.0, literal_drop_p = 0.0, batch_size=100, incremental=True):
 		self.number_of_clauses = number_of_clauses
 		self.number_of_state_bits_ta = number_of_state_bits_ta
 		self.number_of_state_bits_ind = number_of_state_bits_ind
@@ -61,6 +61,10 @@ class TMBasis():
 		self.clause_drop_p = clause_drop_p
 		self.literal_drop_p = literal_drop_p
 
+		self.batch_size = batch_size
+
+		self.incremental = incremental
+
 		self.X_train = np.zeros(0, dtype=np.uint32)
 		self.X_test = np.zeros(0, dtype=np.uint32)
 
@@ -101,8 +105,8 @@ class TMBasis():
 		return self.clause_bank.set_ta_state(clause, ta, state)
 
 class TMClassifier(TMBasis):
-	def __init__(self, number_of_clauses, T, s, confidence_driven_updating=False, type_i_ii_ratio = 1.0, type_iii_feedback=False, d=200.0, platform='CPU', patch_dim=None, feature_negation=True, boost_true_positive_feedback=1, max_included_literals=None, number_of_state_bits_ta=8, number_of_state_bits_ind=8, weighted_clauses=False, clause_drop_p = 0.0, literal_drop_p = 0.0):
-		super().__init__(number_of_clauses, T, s, confidence_driven_updating=confidence_driven_updating, type_i_ii_ratio = type_i_ii_ratio, type_iii_feedback=type_iii_feedback, d=d, platform=platform, patch_dim=patch_dim, feature_negation=feature_negation, boost_true_positive_feedback=boost_true_positive_feedback, max_included_literals=max_included_literals, number_of_state_bits_ta=number_of_state_bits_ta, number_of_state_bits_ind=number_of_state_bits_ind, weighted_clauses=weighted_clauses, clause_drop_p = clause_drop_p, literal_drop_p = literal_drop_p)
+	def __init__(self, number_of_clauses, T, s, confidence_driven_updating=False, type_i_ii_ratio = 1.0, type_iii_feedback=False, d=200.0, platform='CPU', patch_dim=None, feature_negation=True, boost_true_positive_feedback=1, max_included_literals=None, number_of_state_bits_ta=8, number_of_state_bits_ind=8, weighted_clauses=False, clause_drop_p = 0.0, literal_drop_p = 0.0, batch_size=100, incremental=True):
+		super().__init__(number_of_clauses, T, s, confidence_driven_updating=confidence_driven_updating, type_i_ii_ratio = type_i_ii_ratio, type_iii_feedback=type_iii_feedback, d=d, platform=platform, patch_dim=patch_dim, feature_negation=feature_negation, boost_true_positive_feedback=boost_true_positive_feedback, max_included_literals=max_included_literals, number_of_state_bits_ta=number_of_state_bits_ta, number_of_state_bits_ind=number_of_state_bits_ind, weighted_clauses=weighted_clauses, clause_drop_p = clause_drop_p, literal_drop_p = literal_drop_p, batch_size=batch_size, incremental=incremental)
 
 	def initialize(self, X, Y):
 		self.number_of_classes = int(np.max(Y) + 1)
@@ -114,7 +118,7 @@ class TMClassifier(TMBasis):
 		self.clause_banks = []
 		if self.platform == 'CPU':
 			for i in range(self.number_of_classes):
-				self.clause_banks.append(ClauseBank(X, self.number_of_clauses, self.number_of_state_bits_ta, self.number_of_state_bits_ind, self.patch_dim))
+				self.clause_banks.append(ClauseBank(X, self.number_of_clauses, self.number_of_state_bits_ta, self.number_of_state_bits_ind, self.patch_dim, batch_size=self.batch_size, incremental=self.incremental))
 		elif self.platform == 'CUDA':
 			from tmu.clause_bank_cuda import ClauseBankCUDA
 			for i in range(self.number_of_classes):
@@ -215,7 +219,7 @@ class TMClassifier(TMBasis):
 				self.clause_banks[not_target].type_iii_feedback(update_p, self.d, clause_active[not_target]*self.positive_clauses, literal_active, self.encoded_X_train, e, 0)
 		return
 
-	def predict(self, X, incremental=False):
+	def predict(self, X):
 		if not np.array_equal(self.X_test, X):
 			self.encoded_X_test = self.clause_banks[0].prepare_X(X)
 			self.X_test = X.copy()
@@ -225,10 +229,7 @@ class TMClassifier(TMBasis):
 			max_class_sum = -self.T
 			max_class = 0
 			for i in range(self.number_of_classes):
-				if incremental:
-					class_sum = np.dot(self.weight_banks[i].get_weights(), self.clause_banks[i].calculate_clause_outputs_incremental_predict(self.encoded_X_test, e)).astype(np.int32)
-				else:
-					class_sum = np.dot(self.weight_banks[i].get_weights(), self.clause_banks[i].calculate_clause_outputs_predict(self.encoded_X_test, e)).astype(np.int32)
+				class_sum = np.dot(self.weight_banks[i].get_weights(), self.clause_banks[i].calculate_clause_outputs_predict(self.encoded_X_test, e)).astype(np.int32)
 				class_sum = np.clip(class_sum, -self.T, self.T)
 				if class_sum > max_class_sum:
 					max_class_sum = class_sum
@@ -584,25 +585,6 @@ class TMAutoEncoder(TMBasis):
 				self.clause_bank.type_iii_feedback(update_p, self.d, clause_active*(self.weight_banks[target_output].get_weights() >= 0), literal_active, encoded_X, 0, 0)
 		return
 
-	def produce_example(self, the_class, X_csc, X_csr):
-		if self.output_balancing:
-			target = np.random.choice(2)
-			if target == 1:
-				target_indices = X_csc[:,self.output_active[the_class]].indices
-			else:
-				target_indices = np.setdiff1d(np.random.choice(X_csr.shape[0], size=self.accumulation*10, replace=True), X_csc[:,self.output_active[the_class]].indices)
-				while target_indices.shape[0] == 0:
-					target_indices = np.setdiff1d(np.random.choice(X_csr.shape[0], size=self.accumulation*10, replace=True), X_csc[:,self.output_active[the_class]].indices)
-
-			examples = np.random.choice(target_indices, size=self.accumulation, replace=True)
-		else:
-			examples = np.random.choice(X_csr.shape[0], replace=True)
-			target = X_csr[examples,self.output_active[i]]
-
-		accumulated_X = (X_csr[examples].toarray().sum(axis=0) > 0).astype(np.uint32)
-
-		return (target, self.clause_bank.prepare_X(accumulated_X.reshape((1,-1))))
-
 	def activate_clauses(self):
 		# Clauses are dropped based on their weights
 		clause_active = np.ones(self.number_of_clauses, dtype=np.uint32)
@@ -637,7 +619,7 @@ class TMAutoEncoder(TMBasis):
 			self.initialized = True
 
 		X_csr = csr_matrix(X.reshape(X.shape[0], -1))
-		X_csc = csc_matrix(X.reshape(X.shape[0], -1))
+		X_csc = csc_matrix(X.reshape(X.shape[0], -1)).sorted_indices()
 
 		clause_active = self.activate_clauses()
 		literal_active = self.activate_literals()
@@ -652,8 +634,10 @@ class TMAutoEncoder(TMBasis):
 			average_absolute_weights /= self.number_of_classes
 			update_clause = np.random.random(self.number_of_clauses) <= (self.T - np.clip(average_absolute_weights, 0, self.T))/self.T
 
+			Xu, Yu = self.clause_bank.prepare_autoencoder_examples(X_csr, X_csc, self.output_active, self.accumulation)
 			for i in class_index:
-				(target, encoded_X) = self.produce_example(i, X_csc, X_csr)
+				(target, encoded_X) = Yu[i], Xu[i].reshape((1,-1))
+
 				ta_chunk = self.output_active[i] // 32
 				chunk_pos = self.output_active[i] % 32
 				copy_literal_active_ta_chunk = literal_active[ta_chunk]
@@ -693,7 +677,7 @@ class TMAutoEncoder(TMBasis):
 
 	def clause_precision(self, the_class, positive_polarity, X, number_of_examples=2000):
 		X_csr = csr_matrix(X.reshape(X.shape[0], -1))
-		X_csc = csc_matrix(X.reshape(X.shape[0], -1))
+		X_csc = csc_matrix(X.reshape(X.shape[0], -1)).sorted_indices()
 
 		true_positive = np.zeros(self.number_of_clauses, dtype=np.uint32)
 		false_positive = np.zeros(self.number_of_clauses, dtype=np.uint32)
@@ -704,7 +688,8 @@ class TMAutoEncoder(TMBasis):
 		literal_active = self.activate_literals()
 
 		for e in range(number_of_examples):
-			(target, encoded_X) = self.produce_example(the_class, X_csc, X_csr)
+			Xu, Yu = self.clause_bank.prepare_autoencoder_examples(X_csr, X_csc, np.array([self.output_active[the_class]], dtype=np.uint32), self.accumulation)
+			(target, encoded_X) = Yu[0], Xu[0].reshape((1,-1))
 			
 			clause_outputs = self.clause_bank.calculate_clause_outputs_predict(encoded_X, 0)
 
@@ -723,15 +708,16 @@ class TMAutoEncoder(TMBasis):
 
 	def clause_recall(self, the_class, positive_polarity, X, number_of_examples=2000):
 		X_csr = csr_matrix(X.reshape(X.shape[0], -1))
-		X_csc = csc_matrix(X.reshape(X.shape[0], -1))
-
+		X_csc = csc_matrix(X.reshape(X.shape[0], -1)).sorted_indices()
+		
 		true_positive = np.zeros(self.number_of_clauses, dtype=np.uint32)
 		false_negative = np.zeros(self.number_of_clauses, dtype=np.uint32)
 
 		weights = self.weight_banks[the_class].get_weights()
 
 		for e in range(number_of_examples):
-			(target, encoded_X) = self.produce_example(the_class, X_csc, X_csr)
+			Xu, Yu = self.clause_bank.prepare_autoencoder_examples(X_csr, X_csc, np.array([self.output_active[the_class]], dtype=np.uint32), self.accumulation)
+			(target, encoded_X) = Yu[0], Xu[0].reshape((1,-1))
 
 			clause_outputs = self.clause_bank.calculate_clause_outputs_predict(encoded_X, 0)
 
